// rccar_receiver_bluetooth.ino
//
// RCカー を作る （ バギー工作基本セット + Arduino + Bluetooth + Androidスマートフォン )
// 
// Written by Nobuki HIRAMINE
// Last modified : 2012/01/30

#include <SoftwareSerial.h>
#include <PWMServo.h>

#define SERVO_PWM_PIN		(9)
#define BLUETOOTH_RX_PIN	(7)
#define BLUETOOTH_TX_PIN	(6)
#define MOTOR_IN1_PIN		(11)
#define MOTOR_IN2_PIN		(5)

#define READBUFFERSIZE (256)
#define DELIMITER	(",*")

#define MOTORMAX	(200)
#define ANGLEMAX	(30)

SoftwareSerial g_bluetooth( BLUETOOTH_RX_PIN, BLUETOOTH_TX_PIN );
PWMServo g_servo;
char g_szReadBuffer[READBUFFERSIZE] = "";
int  g_iIndexChar = 0;

void setup()
{
	pinMode(MOTOR_IN1_PIN, OUTPUT);
	pinMode(MOTOR_IN2_PIN, OUTPUT);

	g_bluetooth.begin(9600);
	Serial.begin( 9600 );
	g_servo.attach(SERVO_PWM_PIN);
}

int Map( int iIn, int iIn1, int iIn2, int iOut1, int iOut2, boolean bConstrain = false )
{
	double dValue = (double)(iIn - iIn1) * (iOut2 - iOut1) / (iIn2 - iIn1) + iOut1;
	int iValue = (0 < dValue) ? (int)(dValue + 0.5) : (int)(dValue - 0.5);
	if( bConstrain )
	{
		int iOutMin, iOutMax;
		if( iOut1 < iOut2 )
		{
			iOutMin= iOut1;
			iOutMax= iOut2;
		}
		else
		{
			iOutMin= iOut2;
			iOutMax= iOut1;
		}
		if( iOutMin > iValue )
		{
			return iOutMin;
		}
		if( iOutMax < iValue )
		{
			return iOutMax;
		}
	}
	return iValue;
}

void MotorDrive( int iIn1Pin, int iIn2Pin, int iMotor )
{
	if( -5 < iMotor
	 &&  5 > iMotor )
	{
		//digitalWrite(iIn1Pin, HIGH);
		//digitalWrite(iIn2Pin, HIGH);
		digitalWrite(iIn1Pin, LOW);
		digitalWrite(iIn2Pin, LOW);
	}
	else if( 0 < iMotor )
	{
		analogWrite(iIn1Pin, iMotor);
		analogWrite(iIn2Pin, 0);
	}
	else
	{
		analogWrite(iIn1Pin, 0);
		analogWrite(iIn2Pin, -iMotor);
	}
}

// センテンスの解析。
// $HIRARCの場合、引数変数に、モーター、角度を入れ、戻り値 1 を返す。
// $HIRAR以外の場合や、不十分なデータの場合、戻り値は 0 を返す。
int AnalyzeLineString( char szLineString[],
					   boolean& rbMotor, int& riMotor,
					   boolean& rbAngle, int& riAngle )
{
	rbMotor = false;
	rbAngle = false;
	riMotor = 0;
	riAngle = 0;

	// $HIRARC
	if( 0 != strncmp( "$HIRARC", szLineString, 7 ) )
	{
		return 0;
	}
	
	// $HIRARC,10,-20*2D
	char* ptr = szLineString;
	// $HIRARC,10,-20*2D
	ptr = strpbrk( ptr, DELIMITER);	// $HIRARC
	if( NULL == ptr ){ return 0; }
	*ptr = '\0'; ptr++;
	char* pszThrottle = ptr;
	ptr = strpbrk( ptr, DELIMITER);	// スロットル
	if( NULL == ptr ){ return 0; }
	*ptr = '\0'; ptr++;
	char* pszSteering = ptr;
	ptr = strpbrk( ptr, DELIMITER);	// ステアリング
	if( NULL == ptr ){ return 0; }
	*ptr = '\0'; ptr++;
	char* pszChecksum = ptr;
	ptr = strpbrk( ptr, DELIMITER);	// チェックサム
	if( NULL != ptr ){ return 0; }

	if( '\0' != pszThrottle[0] )
	{
		rbMotor = true;
		riMotor = atol(pszThrottle) / 100.0 * MOTORMAX;
	}
	if( '\0' != pszSteering[0] )
	{
		rbAngle = true;
		riAngle = 90 - atol(pszSteering) / 100.0 * ANGLEMAX;
	}

	return 1;
}

// １行文字列の読み込み
// 0 : 読み取り途中。1 : 読み取り完了。
int ReadLineString(	SoftwareSerial& serial,
					char szReadBuffer[], const int ciReadBufferSize, int& riIndexChar,
					char szLineString[], const int ciLineStringSize )
{
	while( 1 )
	{
		char c = serial.read();
		if( -1 == c )
		{
			break;
		}
		if( '\r' == c  )
		{	// 終端
			szReadBuffer[riIndexChar] = '\0';
			strncpy( szLineString, szReadBuffer, ciLineStringSize - 1 );
			szLineString[ciLineStringSize - 1] = '\0';
			riIndexChar = 0;
			return 1;
		}
		else if( '\n' == c )
		{
			;	// 何もしない
		}
		else
		{	// 途中
			if( (ciReadBufferSize - 1) > riIndexChar )
			{
				szReadBuffer[riIndexChar] = c;
				riIndexChar++;
			}
		}
	}

	return 0;
}

void loop()
{
	static int s_iMotor = 0;
	static int s_iAngle = 90;
	
	char szLineString[READBUFFERSIZE];
	if( !ReadLineString( g_bluetooth,
						 g_szReadBuffer, READBUFFERSIZE, g_iIndexChar,
						 szLineString, READBUFFERSIZE ) )
	{	// 読み取り途中
		return;
	}
	// 読み取り完了

	boolean bMotor = false;
	boolean bAngle = false;
	int iMotor = 0;
	int iAngle = 0;
	if( !AnalyzeLineString( szLineString,
							bMotor, iMotor,
							bAngle, iAngle ) )
	{
		return;
	}
	// スロットル、ステアリングを読み取れた

	// スロットル
	if( bMotor )
	{
		s_iMotor = iMotor;
	}
	MotorDrive( MOTOR_IN1_PIN, MOTOR_IN2_PIN, s_iMotor );
	// ステアリング
	if( bAngle )
	{
		s_iAngle =  iAngle;
	}
	g_servo.write( s_iAngle );
}
